/******************************************************************************
    Atualizando a SPIFFS e a FLASH do ESP32 via OTA com WebServer assíncrono
                        Criado em 18 de Novembro de 2022
                                por Michel Galvão

  Eletrogate - Loja de Arduino \\ Robótica \\ Automação \\ Apostilas \\ Kits
                            https://www.eletrogate.com/
******************************************************************************/

//Inclusão das bibliotecas
#include <Adafruit_BMP280.h>
#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <FS.h>
#include <SPIFFS.h>
#include <Update.h>
#include "credenciais.h" // contém as credenciais de acesso ao WiFi

//Definição do endereço I2C do sensor BMP280
#define endereco_bmp280 0x76

//Definição da versão atual do Firmware do ESP32
#define versao_firmware "v1.0"

const unsigned int tempoLimiteConexaoWiFi = 15; // tempo limite da conexao WiFi em segundos

// Instanciação dos objetos das classes das bibliteecas
Adafruit_BMP280 bmp280; // para sensor bmp2280
AsyncWebServer server(80); // para servidor web assíncrono na porta 80

// Função que processa modelos (variáveis) nas páginas web especificadas
String processor(const String& var) {
  if (var == "temperatura")
    return (String)bmp280.readTemperature(); // retorna a temperatura lida do sensor bmp280 em graus Celsius
  if (var == "pressao")
    return (String)(bmp280.readPressure() / 100); // retorna a pressão atmosférica lida do sensor bmp280 em hectopascal
  if (var == "altitude")
    return (String)bmp280.readAltitude(); // retorna a altitude aproximada lida do sensor bmp280 em metros
  if (var == "versao_firmware")
    return (String)versao_firmware; // retorna a versão atual do firmware do ESP32
  return String(); // retorna String vazia
}

bool deveReiniciar = false; // variável que controla se o ESP32 deve ou não reiniciar

void setup() {
  Serial.begin(115200); // configura a taxa de transferência em 115200 bits por segundo (baud rate) para a transmissão serial
  Serial.println("Firmware iniciado"); // informa ao usuário que o firmware foi iniciado

  if (!bmp280.begin(endereco_bmp280)) // se a inicialização do sensor BMP280 no endereço contido na variável não for bem sucedido, ...
  {
    Serial.println("Não foi possível inicializar o sensor BMP280. Verifique as conexões elétricas."); // informa ao usuário que não foi possível inicializar o sensor BMP280
    return; // para o programa
  }

  delay(200);

  WiFi.disconnect(); // desconecta de qualquer conexão anterior, esta isntrução evita problemas de conexão
  WiFi.mode(WIFI_STA); // configura o WiFi para o modo de estação
  WiFi.begin(ssid, password); // Conecta na rede WiFi com as credenciais informadas.

  Serial.print("Fazendo tentativa de conexão ao WiFi...");
  int qtdTentativasConexao = 0;
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
    qtdTentativasConexao++;

    if (qtdTentativasConexao > tempoLimiteConexaoWiFi * 2) {
      Serial.println("\nTempo limite de Conexão WiFi falhada. Programa finalizado."); // informa ao usuário que não foi possível se conectar ao WiFi
      while (1);
    }
  }

  Serial.println("\nConexão ao WiFi foi bem sucedida");

  Serial.print("Endereço IP: ");
  Serial.println(WiFi.localIP()); // informa ao usuário o endereço IP ao qual o ESP32 recebeu da rede WiFi

  if (!SPIFFS.begin(true)) // se a montagem do sistema de arquivos SPIFFS (com a opção de formatar ativada em caso de falhas), ...
  {
    Serial.println("SPIFFS não inicializada"); // informa ao usuário que a partição SPIFFS não foi bem inicializada
    return; // para o programa
  }

  // disponibiliza o url "/"
  server.on("/", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->redirect("/index.html"); // redireciona para página index.html
  });

  // disponibiliza o url "/index.html"
  server.on("/index.html", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, // o tipo do Filesystem usado é o SPIFFS
                  "/index.html", // o caminho do arquivo no diretório do Filesystem SPIFFS é /index.html
                  "text/html", // o tipo de mídia da Internet (Media type) é HTML
                  false, // o arquivo não vai ser disponibilizado para download direto
                  processor // parâmetro para processar conteúdo contendo modelos
                 ); // responde à solicitação do cliente enviando-lhe a página index.html.
  });

  // disponibiliza o url "/atualizar-firmware.html"
  server.on("/atualizar-firmware.html", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, "/atualizar-firmware.html", // o caminho do arquivo no diretório do Filesystem SPIFFS é /atualizar-firmware.html
                  "text/html", false, processor); // responde à solicitação do cliente enviando-lhe a página atualizar-firmware.html.
  });

  // disponibiliza o url "/atualizar-firmware.js"
  server.on("/atualizar-firmware.js", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, "/atualizar-firmware.js", // o caminho do arquivo no diretório do Filesystem SPIFFS é /atualizar-firmware.js
                  "text/javascript");// responde à solicitação do cliente, sem processar modelos da página, enviando-lhe a página atualizar-firmware.js
  });

  // disponibiliza o url "/index.css"
  server.on("/index.css", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, "/index.css", // o caminho do arquivo no diretório do Filesystem SPIFFS é /index.css
                  "text/css");// responde à solicitação do cliente, sem processar modelos da página, enviando-lhe a index.css
  });

  // disponibiliza o url "/atualizar-firmware.css"
  server.on("/atualizar-firmware.css", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, "/atualizar-firmware.css", // o caminho do arquivo no diretório do Filesystem SPIFFS é /atualizar-firmware.css
                  "text/css");// responde à solicitação do cliente, sem processar modelos da página, enviando-lhe a página script-atualizar-firmware.css
  });

  // disponibiliza o url "/atualizar-flash"
  server.on("/atualizar-flash", HTTP_POST,
  [] (AsyncWebServerRequest * request) {
    deveReiniciar = !Update.hasError();
    AsyncWebServerResponse *response = request->beginResponse(200, "text/plain", deveReiniciar ? "OK" : "FALHO");
    response->addHeader("Connection", "close");
    request->send(response);
  },
  [](AsyncWebServerRequest * request, String filename, size_t index, uint8_t *data, size_t len, bool final) {
    if (!index) {
      Serial.println("Atualização de FLASH iniciada");
      Serial.print("Nome do Arquivo: ");
      Serial.print(filename);

      if (!Update.begin(UPDATE_SIZE_UNKNOWN)) {
        Update.printError(Serial);
      }
    }
    if (!Update.hasError()) {
      /* piscando firmware para ESP32 */
      if (Update.write(data, len) != len) {
        Update.printError(Serial);
      }
    }
    if (final) {
      if (Update.end(true)) { // true para definir o tamanho para o progresso atual
        Serial.print("Atualização efetuada com sucesso: ");
        Serial.print(index + len);
        Serial.println(" bytes");
        Serial.println("Reiniciando...\n");
      } else {
        Update.printError(Serial);
      }
    }
  });

  server.on("/atualizar-spiffs", HTTP_POST,
  [] (AsyncWebServerRequest * request) {
    deveReiniciar = !Update.hasError();
    Serial.println("Conexão fechada");
    AsyncWebServerResponse *response = request->beginResponse(200, "text/plain", deveReiniciar ? "OK" : "FALHO");
    response->addHeader("Connection", "close");
    request->send(response);
  },
  [](AsyncWebServerRequest * request, String filename, size_t index, uint8_t *data, size_t len, bool final) {

    if (!index) {
      Serial.println("Atualização de SPIFFS iniciada");
      Serial.print("Nome do Arquivo: ");
      Serial.print(filename);

      if (!Update.begin(UPDATE_SIZE_UNKNOWN, U_SPIFFS)) {
        Update.printError(Serial);
      }
    }
    if (!Update.hasError()) {
      /* piscando firmware para ESP32 */
      if (Update.write(data, len) != len) {
        Update.printError(Serial);
      }
    }
    if (final) {
      if (Update.end(true)) { // true para definir o tamanho para o progresso atual
        Serial.print("Atualização efetuada com sucesso: ");
        Serial.print(index + len);
        Serial.println(" bytes");
        Serial.println("Programado para Reinicio...\n");
      } else {
        Update.printError(Serial);
      }
    }
  });

  // Servidor começa à ouvir os clientes
  server.begin();


}

void loop() {
  if (isnan(bmp280.readAltitude()))
  {
    Serial.println("Erro na leitura do sensor BMP280");
    delay(1000);
  }
  if (deveReiniciar) // verifica se o ESP32 deve ou não reiniciar
  {
    Serial.println("Reiniciando ESP32...");
    delay(100);
    ESP.restart(); // reinicia o ESP32
  }
  delay(1);
}
